import os
import typing as t
from datetime import datetime
from decimal import Context
from functools import reduce

from lifeomic_chatbot_tools._utils import ImportExtraError


try:
    import boto3
    from boto3.dynamodb.conditions import Key
    from botocore.config import Config
except ImportError:
    raise ImportExtraError("aws", __name__)

from lifeomic_chatbot_tools.persistence.record_store.base import IRecord, IRecordStore


KeyDict = t.Dict[str, t.Any]
"""
The type of primary key/unique record identifier that DynamoDB uses, since it can sometimes use composite keys, which
are a combination of the partition key and sort key.
"""


class IDynamoDBRecord(IRecord[KeyDict], t.Protocol):
    def dict(self) -> dict:
        """Converts this object into a dictionary that DynamoDB can save."""
        ...

    @classmethod
    def parse_obj(cls, obj: t.Any) -> "IDynamoDBRecord":
        """Parses a representation of this class produced by `self.dict(...)` into a class object."""
        ...


TIDynamoDBRecord = t.TypeVar("TIDynamoDBRecord", bound=IDynamoDBRecord)


class DynamoDBRecordStore(t.Generic[TIDynamoDBRecord], IRecordStore[KeyDict, TIDynamoDBRecord]):
    """
    A DynamoDB DAO for Pydantic data models. In addition to its parent class's `WHERE equals` behavior, this class
    also supports arbitrary `WHERE` clauses (e.g. ``<=``), using Firestore's `operator string syntax <https://googleapis
    .dev/python/firestore/latest/collection.html?highlight=where#google.cloud.firestore_v1.base_collection.BaseCollectio
    nReference.where>`_.
    """

    def __init__(
        self,
        table_name: str,
        record_class: t.Type[TIDynamoDBRecord],
        *,
        primary_key_field_name="id",
        sort_key_field_name=None,
    ):
        self.record_cls = record_class
        self._table = boto3.resource(
            "dynamodb", endpoint_url=os.getenv("AWS_ENDPOINT"), config=Config(region_name=os.getenv("AWS_REGION"))
        ).Table(table_name)
        self._pk = primary_key_field_name
        self._sk = sort_key_field_name
        self._decimal_ctx = Context(prec=38)  # dynamodb has a maximum precision of 38 digits for decimal numbers

    def save(self, record: TIDynamoDBRecord):
        item_dict = {**record.dict(), **record.get_id()}
        item_dict = self.make_dynamodb_compatible(item_dict)
        self._table.put_item(Item=item_dict)

    def get(self, id_: KeyDict) -> t.Optional[TIDynamoDBRecord]:
        res = self._table.get_item(Key=self.make_dynamodb_compatible(id_))
        if "Item" not in res:
            return None
        return t.cast(TIDynamoDBRecord, self.record_cls.parse_obj(res["Item"]))

    def delete(self, id_: KeyDict) -> bool:
        res = self._table.delete_item(Key=self.make_dynamodb_compatible(id_), ReturnValues="ALL_OLD")
        # `ReturnValues="ALL_OLD"` causes DynamoDB to return the attributes of the item as it appeared
        # before the DeleteItem operation. If there are no attributes, we assume the item did not exist
        # before the delete operation, and so was not actually deleted.
        return len(res.get("Attributes", {})) > 0

    def delete_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> int:
        """
        Deletes all records which satisfy the optional ``*conditions`` and ``*where_equals`` conditions. Returns the
        number of records that were deleted. **Note**: the current implementation for this method is very slow, using
        the DynamoDB scan method under the hood.
        """
        # TODO: This is an incredibly slow way of doing this. Use indexes and a batch delete instead.
        num_deleted = 0
        for record in self.get_all(*conditions, **where_equals):
            self.delete(record.get_id())
            num_deleted += 1
        return num_deleted

    def get_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> t.Iterable[TIDynamoDBRecord]:
        """
        Paginates over all records in the table, yielding them in an iterator. Retrieves all records which satisfy the
        optional ``*conditions`` and ``**where_equals`` equality conditions.
        """
        done, start_key, use_query = False, None, False
        conditions_expression = self._set_conditions(*conditions, **where_equals)
        if conditions_expression is None:
            conditions_expression = {}
        if "KeyConditionExpression" in conditions_expression.keys():
            use_query = True
        while not done:
            if start_key:
                if use_query:
                    res = self._table.query(ExclusiveStartKey=start_key, **conditions_expression)
                else:
                    res = self._table.scan(ExclusiveStartKey=start_key, **conditions_expression)
            else:
                if use_query:
                    res = self._table.query(**conditions_expression)
                else:
                    res = self._table.scan(**conditions_expression)
            start_key = res.get("LastEvaluatedKey")
            done = start_key is None
            for item in res.get("Items", []):
                yield t.cast(TIDynamoDBRecord, self.record_cls.parse_obj(item))

    def _set_conditions(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> dict:
        """
        `FilterExpression` specifies a condition that returns only items that satisfy the condition - all other items
        are discarded. However, the filter is applied only after the entire table has been scanned.
        `KeyConditionExpression` specifies a condition for partition key or sort key. It requires an equality check on a
        partition key value, and optional other operators check on sort key. Using `KeyConditionExpression`, `Query`
        performs a direct lookup to a selected partition based on primary or secondary partition/hash. `Scan` only
        accepts `FilterExpression` argument, where as query accepts both `KeyConditionExpression`, and
        `FilterExpression`. Important: To use `Query`, we must provide `KeyConditionExpression` --> so if
        `KeyConditionExpression` is not provided, we need to use `Scan`, otherwise we use `Query`.
        """
        if self._sk is None:
            filters_list, p_key_flag, p_key_condition = [], False, None
            for attr, value in where_equals.items():
                if attr == self._pk and p_key_flag is False:
                    p_key_flag = True
                    p_key_condition = Key(attr).eq(self.make_dynamodb_compatible(value))
                else:
                    filters_list.append(Key(attr).eq(self.make_dynamodb_compatible(value)))
            for attr, op, value in conditions:
                if attr == self._pk and op == "==" and p_key_flag is False:
                    p_key_flag = True
                    p_key_condition = Key(attr).eq(self.make_dynamodb_compatible(value))
                else:
                    filters_list.append(self.choose_operator(attr, op, value))

            if len(filters_list) == 0 and p_key_flag is False:
                return {}
            if len(filters_list) == 0 and p_key_flag:
                return {"KeyConditionExpression": p_key_condition}

            if len(filters_list) != 0 and p_key_flag is False:
                res = reduce(lambda x, y: x & y, filters_list)
                return {"FilterExpression": res}
            if len(filters_list) != 0 and p_key_flag:
                res = reduce(lambda x, y: x & y, filters_list)
                return {"KeyConditionExpression": p_key_condition, "FilterExpression": res}

        else:
            filters_list, p_key_flag, sort_key_list, p_key_condition = [], False, [], None
            for attr, value in where_equals.items():
                if attr == self._pk and p_key_flag is False:
                    p_key_flag = True
                    p_key_condition = Key(attr).eq(self.make_dynamodb_compatible(value))
                elif attr == self._sk:
                    sort_key_list.append(Key(attr).eq(self.make_dynamodb_compatible(value)))
                else:
                    filters_list.append(Key(attr).eq(self.make_dynamodb_compatible(value)))
            for attr, op, value in conditions:
                if attr == self._pk and op == "==" and p_key_flag is False:
                    p_key_flag = True
                    p_key_condition = Key(attr).eq(self.make_dynamodb_compatible(value))
                elif attr == self._sk:
                    sort_key_list.append(self.choose_operator(attr, op, value))
                else:
                    filters_list.append(self.choose_operator(attr, op, value))

            # important check
            if p_key_flag and len(sort_key_list) >= 3:
                raise Exception("keyconditionexpressions must only contain one condition per key")
            if p_key_flag and len(sort_key_list) == 2:
                # it converts (a <= x <= b) into Key(x).between(a, b),
                # pay attention that keyconditionexpressions must only contain one condition per key
                sort_key_list = self.combine_multiple_condition_sort_key(*conditions, **where_equals)

            if len(filters_list) == 0 and p_key_flag is False and len(sort_key_list) == 0:
                return {}
            if len(filters_list) == 0 and p_key_flag and len(sort_key_list) == 0:
                return {"KeyConditionExpression": p_key_condition}
            if len(filters_list) == 0 and p_key_flag and len(sort_key_list) != 0:
                sort_key_list.append(p_key_condition)
                key_res = reduce(lambda x, y: x & y, sort_key_list)
                return {"KeyConditionExpression": key_res}
            if len(filters_list) == 0 and p_key_flag is False and len(sort_key_list) != 0:
                res = reduce(lambda x, y: x & y, sort_key_list)
                return {"FilterExpression": res}

            if len(filters_list) != 0 and p_key_flag is False and len(sort_key_list) == 0:
                res = reduce(lambda x, y: x & y, filters_list)
                return {"FilterExpression": res}
            if len(filters_list) != 0 and p_key_flag and len(sort_key_list) == 0:
                res = reduce(lambda x, y: x & y, filters_list)
                return {"KeyConditionExpression": p_key_condition, "FilterExpression": res}
            if len(filters_list) != 0 and p_key_flag and len(sort_key_list) != 0:
                sort_key_list.append(p_key_condition)
                key_res = reduce(lambda x, y: x & y, sort_key_list)
                res = reduce(lambda x, y: x & y, filters_list)
                return {"KeyConditionExpression": key_res, "FilterExpression": res}
            if len(filters_list) != 0 and p_key_flag is False and len(sort_key_list) != 0:
                filters_list.extend(sort_key_list)
                res = reduce(lambda x, y: x & y, filters_list)
                return {"FilterExpression": res}
        return {}

    def choose_operator(self, attr, op, value):
        value = self.make_dynamodb_compatible(value)
        if op == "<=":
            return Key(attr).lte(value)
        if op == "<":
            return Key(attr).lt(value)
        if op == ">=":
            return Key(attr).gte(value)
        if op == ">":
            return Key(attr).gt(value)
        if op == "==":
            return Key(attr).eq(value)

    def combine_multiple_condition_sort_key(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> list:
        lte_flag, gte_flag, sort_key_list, dict_value = False, False, [], {}
        for attr, value in where_equals.items():
            if attr == self._sk:
                sort_key_list.append(Key(attr).eq(self.make_dynamodb_compatible(value)))
                return sort_key_list
        for attr, op, value in conditions:
            if attr == self._sk and op == "==":
                sort_key_list.append(Key(attr).eq(self.make_dynamodb_compatible(value)))
                return sort_key_list
            if attr == self._sk and op == "<=":
                lte_flag = True
                dict_value["lte"] = self.make_dynamodb_compatible(value)
            if attr == self._sk and op == ">=":
                gte_flag = True
                dict_value["gte"] = self.make_dynamodb_compatible(value)
        if lte_flag and gte_flag:
            sort_key_list.append(Key(self._sk).between(dict_value["gte"], dict_value["lte"]))
            return sort_key_list
        return []

    def remap(self, o: object, apply: t.Callable):
        if isinstance(o, dict):
            return {k: self.remap(v, apply) for k, v in o.items()}
        elif isinstance(o, (list, tuple, set)):
            return type(o)(self.remap(elem, apply) for elem in o)
        elif isinstance(o, (str, int, float, type(None), datetime)):
            return apply(o)
        else:
            raise AssertionError(f"remap encountered unsupported type {type(o)}")

    def make_value_dynamodb_compatible(self, value):
        """Tries to convert ``value`` to a type DynamoDB can handle."""
        if isinstance(value, datetime):
            # An attribute with the type of datetime is recorded as iso format, so the value being used for conditional
            # search should be iso format as well.
            value = value.isoformat()
        elif isinstance(value, float):
            # DynamodB doesn't support float type. It must be converted to decimal. Also, a local `decimal.Context` is
            # needed because DynamoDB's default context will throw an error if any rounding occurs.
            value = self._decimal_ctx.create_decimal_from_float(value)
        return value

    def make_dynamodb_compatible(self, o: object):
        return self.remap(o, self.make_value_dynamodb_compatible)



import os
import pickle
import shutil
import typing as t


class ITypeSerializer(t.Protocol):
    """
    Interface for serializing instances of some type or group of types, for use with the :class:`Serializer` class.
    """

    type_name: str
    """
    A name identifying the type this serializer serializers. Should be unique among all the other
    :class:`TypeSerializer` serializers used. Should also contain only letters, numbers, and dashes.
    """

    ext: t.Optional[str]
    """The filename extension, if the serializer serializes its data to a single file. Should be ``None`` otherwise."""

    def serialize(self, obj: object, path: str):
        """Should serialize ``obj`` and save it to ``path``."""
        ...

    def deserialize(self, path: str) -> object:
        """
        Should load the persisted object at ``path``, and return the deserialized version. The object living at ``path``
        is guaranteed to have the type associated with this serializer.
        """
        ...

    def is_serializable(self, obj: object) -> bool:
        """Should return ``True`` if this serializer should be used to serialze ``obj``."""
        ...


class _CustomPickler(pickle.Pickler):
    def __init__(self, pkl_file, assets_path: str, type_serializers: t.Sequence[ITypeSerializer]):
        super().__init__(pkl_file)
        self._assets_path = assets_path
        self._ser_map = {ser.type_name: ser for ser in type_serializers}
        self._unique_id = 0

    def persistent_id(self, obj: object) -> t.Optional[tuple]:
        for ser in self._ser_map.values():
            if ser.is_serializable(obj):
                # Treat `obj` as an external object and serialize it using our own
                # methods. Our serializer's type name and the relative path it was serialized
                # to is returned and pickled, so the pickler will know how to find
                # the object again and deserialize it.
                obj_id = f"{ser.type_name}-{self._get_unique_id()}"
                obj_path = self._resolve_path(ser, obj_id)
                ser.serialize(obj, os.path.join(self._assets_path, obj_path))
                return ser.type_name, obj_path

        # No custom serializer for `obj`; pickle it using the normal way.
        return None

    def _get_unique_id(self) -> int:
        """A primary key generator."""
        id_ = self._unique_id
        self._unique_id += 1
        return id_

    @staticmethod
    def _resolve_path(ser: ITypeSerializer, obj_id: str):
        return f"{obj_id}.{ser.ext}" if ser.ext else obj_id


class _CustomUnpickler(pickle.Unpickler):
    def __init__(self, pkl_file, assets_path: str, type_serializers: t.Sequence[ITypeSerializer]):
        super().__init__(pkl_file)
        self._assets_path = assets_path
        self._ser_map = {ser.type_name: ser for ser in type_serializers}

    def persistent_load(self, pid: tuple) -> object:
        """
        This method is invoked whenever a persistent ID is encountered.
        Here, pid is the tuple returned by `_CustomPickler.persistent_id`.
        """
        ser_type_name, obj_path = pid
        if ser_type_name not in self._ser_map:
            raise pickle.UnpicklingError(
                "cannot deserialize: an object was found which was serialized using the "
                f"{ser_type_name} serializer, and this unpickler does not have that serializer registered."
            )
        ser = self._ser_map[ser_type_name]
        return ser.deserialize(os.path.join(self._assets_path, obj_path))


class Serializer:
    """
    A replacement for the :func:`pickle.dump` and :func:`pickle.load` functions. Exposes a hook for custom serialization
    behavior of different data types (e.g. keras models, PyTorch models, numpy arrays, etc.). You can use your own type
    serializers. Just implement the :class:`TypeSerializer` class and pass an instance of your class to the constructor.
    """

    def __init__(self, *custom_type_serializers: ITypeSerializer):
        self._type_serializers = custom_type_serializers

        if len(self._type_serializers) != len({ser.type_name for ser in self._type_serializers}):
            raise ValueError(
                "The type_name of each type serializer must be unique."
                f" Currently registered names: {[ser.type_name for ser in self._type_serializers]}"
            )

    def serialize(self, obj: object, path: str, overwrite: bool = False):
        """Serialize ``obj`` to ``path``, a directory."""
        if os.path.isfile(path):
            raise ValueError(f"cannot serialize to directory {path}; it is a file")
        if not os.path.isdir(path):
            os.makedirs(path)
        elif len(os.listdir(path)) > 0 and not overwrite:
            raise ValueError(f"cannot serialize: {path} is already populated and overwrite is set to False")

        with open(self._get_pkl_path(path), "wb") as f:
            _CustomPickler(f, path, self._type_serializers).dump(obj)

    def deserialize(self, path: str, delete: bool = False):
        """
        Load the data that was serialized to the directory at
        ``path``. It should have been serialized using this class's
        :meth:`serialize` method. If ``delete==True``, ``path`` will be
        deleted once the deserialization is finished.
        """
        # Deserialize the data
        with open(self._get_pkl_path(path), "rb") as f:
            obj = _CustomUnpickler(f, path, self._type_serializers).load()

        if delete:
            shutil.rmtree(path)

        return obj

    @staticmethod
    def _get_pkl_path(path: str) -> str:
        return os.path.join(path, "data.pkl")


class Persistent:
    """
    Mixin class giving persistence behavior. Any inheriting subclass will automatically receive :meth:`to_dir` and
    :meth:`from_dir` methods, which allows a class instance to easily be serialized to and deserialized from a
    directory.
    """

    serializer = Serializer()
    """The serializer to use. Override this for custom serialization behavior."""

    def to_dir(self, path: str, overwrite: bool = False):
        """Serializes the full state of ``self`` to directory ``path``."""
        self.serializer.serialize(self, path, overwrite)

    @classmethod
    def from_dir(cls, path: str, delete: bool = False) -> "Persistent":
        """
        Deserializes a full instance of this class from directory ``path``. If ``delete==True``,
        the persisted instance will be deleted once loaded into memory.
        """
        obj = cls.serializer.deserialize(path, delete)
        assert isinstance(obj, cls)
        return obj



import typing as t
from functools import wraps


class ImportExtraError(ImportError):
    """A needed package extra has not been installed."""

    def __init__(self, extra_name: str, feature_name: str):
        super().__init__(f"The `{extra_name}` package extra is required to use `{feature_name}`.")


def requires_extras(**has_extras: bool):
    def decorator(callable_: t.Callable):
        @wraps(callable_)
        def check_extras(*args, **kwargs):
            for extra_name, has_extra in has_extras.items():
                if not has_extra:
                    raise ImportExtraError(extra_name, callable_.__name__)
            return callable_(*args, **kwargs)

        return check_extras

    return decorator



import hashlib

from pydantic import BaseModel


def hash_model(obj: BaseModel, **kwargs) -> str:
    """
    Creates a deterministic hash of ``obj``, which is some pydantic model. All keyword arguments are forwarded on to
    :meth:`pydantic.BaseModel.json`.
    """
    digest = hashlib.sha256(obj.json(sort_keys=True, **kwargs).encode("utf-8")).hexdigest()
    return digest



import time
import typing as t
from abc import ABC, abstractmethod

from fastapi import HTTPException, status
from loguru import logger

from lifeomic_chatbot_tools.persistence.record_store.base import IRecord, IRecordStore


class IServiceVersionMetadata(IRecord, t.Protocol):
    name: str
    """The name of this version."""

    synced_at: float
    """Time the service version was most recently synced."""


TIServiceVersionMetadata = t.TypeVar("TIServiceVersionMetadata", bound=IServiceVersionMetadata)
TServiceVersionId = t.TypeVar("TServiceVersionId")


class IDatasetRecord(IRecord, t.Protocol):
    """Minimum required interface for a generic dataset object managed by an artifact manager."""

    agent_id: str
    updated_at: float

    @property
    def digest(self) -> str:
        """A hash of this record, which should be invariant to the `updated_at` attribute."""
        ...


TIDatasetRecord = t.TypeVar("TIDatasetRecord", bound=IDatasetRecord)
TDatasetId = t.TypeVar("TDatasetId")


class IArtifactRecord(IRecord, t.Protocol):
    """Minimum required interface for an artifact produced by a versioned dataset and a versioned model."""

    agent_id: str
    dataset_digest: str
    """The hash of the dataset that was used to produce this artifact."""

    service_version: str
    """The version of the model that produced this artifact."""

    updated_at: float


TIArtifactRecord = t.TypeVar("TIArtifactRecord", bound=IArtifactRecord)
TArtifactId = t.TypeVar("TArtifactId")


class BaseArtifactManager(
    t.Generic[TIArtifactRecord, TArtifactId, TIDatasetRecord, TDatasetId, TIServiceVersionMetadata, TServiceVersionId],
    ABC,
):
    """
    Clas which provides a simple API for creating, persisting, and managing artifacts produced by a versioned machine
    learning (ML) model. Each artifact is associated with an agent. An artifact is the deterministic product of a
    specific ML model version, and a specific version of a dataset.

    Attributes
    ----------
    get_version : callable
        A function which takes no arguments and returns the current model version when called.
    """

    def __init__(
        self,
        artifacts: IRecordStore[TArtifactId, TIArtifactRecord],
        datasets: IRecordStore[TDatasetId, TIDatasetRecord],
        versions: IRecordStore[TServiceVersionId, TIServiceVersionMetadata],
        version: t.Union[t.Callable[[], str], str],
        make_artifact_id: t.Callable[[str, str], TArtifactId],
        make_dataset_id: t.Callable[[str], TDatasetId],
        *,
        max_service_versions=5,
    ):
        """
        Parameters
        ----------
        artifacts : IRecordStoreStringId
            Data store used to store the artifacts.
        datasets : IRecordStoreStringId
            Data store used to store the datasets which are used to produce the artifacts.
        versions : IRecordStoreStringId
            Data store used to store metadata about each model version this manager interacts with.
        version : str or callable
            A function which takes no arguments, which should return the current model version when called. This is
            passed in as a function to support the use case where the underlying model version could change at any
            moment, for example because a new model was trained and deployed behind the scenes. If the model version
            does not change, a string can be passed in instead.
        make_artifact_id : callable
            A function which takes a service version and an agent id, and creates an ID for the artifact object created
            under that version, for that agent. We have an artifact's database ID be the deterministic product of the
            service version that created it, and the agent that the artifact is for. We do this because for a given
            service version, we only need to persist one artifact version per agent at a time.
        make_dataset_id : callable
            A function which accepts an agent id, and returns an id for that agent's dataset object, to be used to store
            the dataset in the dataset store.
        max_service_versions : int, optional
            The number of model versions (also called service versions) to track in the ``versions`` data store.
        """
        if isinstance(version, str):
            self.get_version = lambda: version
        else:
            self.get_version = version
        self._max_service_versions = max_service_versions
        self._artifacts = artifacts
        self._datasets = datasets
        self._versions = versions
        self._make_artifact_id = make_artifact_id
        self._make_dataset_id = make_dataset_id

    @abstractmethod
    def create_artifact_from_dataset(self, dataset: TIDatasetRecord) -> TIArtifactRecord:
        """
        Implementing subclasses should take ``dataset``, and the model associated with version ``self.get_version()``,
        and produce an artifact.
        """
        pass

    @abstractmethod
    def create_version_metadata(self, version: str, synced_at: float) -> TIServiceVersionMetadata:
        pass

    def delete_artifact(self, agent_id: str):
        """
        Deletes from the database a dataset and all artifacts associated with it, if they exist. Returns the number of
        total database records that were deleted.
        """
        num_deleted = int(self._datasets.delete(self._make_dataset_id(agent_id)))
        num_deleted += self._artifacts.delete_all(agent_id=agent_id)  # get rid of any old versions as well
        return num_deleted

    def save_artifact(self, artifact: TIArtifactRecord, dataset: TIDatasetRecord):
        """
        Serializes and saves the artifact. Also saves the dataset that was used to create the artifact, so this manager
        can recreate the artifact at any time if needed.
        """
        self._artifacts.save(artifact)
        self._datasets.save(dataset)

    def load_artifact(self, agent_id: str) -> TIArtifactRecord:
        """
        Load an agent's artifact from the database. If the agent's dataset exists but its artifact doesn't, then compute
        it, save it (so its available next time), and then return it.
        """
        dataset = self._datasets.get(self._make_dataset_id(agent_id))
        if dataset is None:
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                f"no dataset exists for agent id {agent_id}; artifact cannot be retrieved or computed",
            )
        version = self.get_version()
        artifact = self._artifacts.get(self._make_artifact_id(version, agent_id))
        if artifact is not None and artifact.dataset_digest == dataset.digest:
            return artifact
        else:
            logger.info(
                f"artifact for agent {agent_id} does not exist for service version {version} and dataset "
                f"digest {dataset.digest}; creating it now"
            )
            # Artifact has not yet been created for this dataset version and service version.
            artifact = self.create_artifact_from_dataset(dataset)
            self.save_artifact(artifact, dataset)
            return artifact

    def sync(self) -> int:
        """
        Ensures this service version has its own artifact for all currently saved datasets. Returns the number of
        artifacts that had to be created for that to happen.
        """
        digest2agent = {dataset.digest: dataset.get_id() for dataset in self._datasets.get_all()}
        all_dataset_digests = set(digest2agent.keys())
        version = self.get_version()
        datasets_currently_indexed = {
            artifact.dataset_digest for artifact in self._artifacts.get_all(service_version=version)
        }
        datasets_to_index = all_dataset_digests - datasets_currently_indexed
        logger.info(
            f"creating artifact for {len(datasets_to_index)}/{len(all_dataset_digests)} "
            f"existing datasets for service version {version}"
        )
        for digest in datasets_to_index:
            dataset = self._datasets.get(digest2agent[digest])
            if dataset is None:
                raise AssertionError(
                    f"Expected dataset to exist for agent {digest2agent[digest]}. It existed just a little bit ago."
                )
            logger.info(f"creating artifact for dataset digest={digest} associated with agent {digest2agent[digest]}")
            artifact = self.create_artifact_from_dataset(dataset)
            self.save_artifact(artifact, dataset)

        metadata = self.create_version_metadata(version, time.time())
        self._versions.save(metadata)
        logger.info("service version sync utility finished successfully.")
        self._remove_old_service_versions()
        return len(datasets_to_index)

    def _remove_old_service_versions(self):
        """
        Removes all artifacts for any old service versions. We only keep data for the ``self._max_service_versions``
        most recent service versions. The old versions are the ones that have been synced least recently.
        """
        versions = list(self._versions.get_all())
        n_to_remove = len(versions) - self._max_service_versions
        if n_to_remove > 0:
            logger.info(f"removing data for {n_to_remove} old service versions")
            # Remove the oldest versions.
            versions_to_remove = sorted(versions, key=lambda v: v.synced_at)[:n_to_remove]
            for version in versions_to_remove:
                self._artifacts.delete_all(service_version=version.name)
                self._versions.delete(version.get_id())



import typing as t

from fastapi import FastAPI, Request, Response
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.routing import APIRoute
from loguru import logger as _logger
from starlette.exceptions import HTTPException as StarletteHTTPException


def _make_error_logging_route_handler_class(*, msg_prefix="", logger) -> t.Type[APIRoute]:
    """
    Makes a FastAPI route handler class which intercepts and logs all internal errors that occur in a FastAPI app.
    Related docs: https://fastapi.tiangolo.com/advanced/custom-request-and-route/
    """

    class ErrorLoggingRouteHandler(APIRoute):
        def get_route_handler(self) -> t.Callable:
            original_route_handler = super().get_route_handler()

            async def custom_route_handler(request: Request) -> Response:
                try:
                    # Try to handle the route as normal.
                    return await original_route_handler(request)
                except Exception as exc:
                    # Some type of exception has occurred. If it's a normal Python exception or a 500-level exception,
                    # log it here.
                    error_to_report = None
                    is_fastapi_error = isinstance(exc, (StarletteHTTPException, RequestValidationError))
                    if not is_fastapi_error:
                        # This is an unknown internal exception.
                        error_to_report = msg_prefix + str(exc)

                    if isinstance(exc, StarletteHTTPException) and exc.status_code >= 500:
                        # This is a 500-level internal exception.
                        error_to_report = msg_prefix + exc.detail

                    if error_to_report is not None:
                        # Log the error.
                        logger.exception(error_to_report)

                    if is_fastapi_error:
                        # FastAPI has a native way for handling these types of errors, so we'll let it bubble up to
                        # those handlers.
                        raise exc
                    else:
                        # Turn the normal Python exception into an HTTP exception with a message for the caller.
                        # This is the error FastAPI will return to the user.
                        server_response = "An internal error has occurred. Our team has been notified of this incident."
                        raise HTTPException(500, server_response)

            return custom_route_handler

    return ErrorLoggingRouteHandler


def with_error_logging(app: FastAPI, msg_prefix="", logger=_logger):
    """
    Adds middleware to ``app`` which intercepts and logs all internal errors that occur in the app. Intercepts all
    ``Exception`` objects which are uncaught by the app, as well as all 500-level errors intentionally thrown by the
    app.

    Parameters
    ----------
    app : FastAPI
        The FastAPI app to add the middleware to.
    msg_prefix : str, optional
        If provided, this string will be pre-pended to the error messages of all intercepted errors before they are
        logged and reported.
    logger : object, optional
        logger to call when generating log lines. Defaults to the default loguru logger. Must have an ``exception``
        method that can be called e.g. ``logger.exception("error")``.
    """
    app.router.route_class = _make_error_logging_route_handler_class(msg_prefix=msg_prefix, logger=logger)
    return app



import typing as t
from collections import defaultdict
from itertools import chain

from loguru import logger
from pydantic import BaseModel

from lifeomic_chatbot_tools.types.conversations.actions import AgentAction
from lifeomic_chatbot_tools.types.conversations.conversation import Conversation, ConversationDataset
from lifeomic_chatbot_tools.types.conversations.dialogue_turns import UserDialogueTurn
from lifeomic_chatbot_tools.types.nlu import NLUExample, NLUExampleDataset


class Intent(BaseModel):
    name: str


class Slot(BaseModel):
    name: str


class AgentConfig(BaseModel):
    """A configuration for a chatbot, including its NLU and dialogue policy training data."""

    name: str
    """The chatbot's name."""

    agentId: t.Optional[str]
    """The chatbot's unique id."""

    accountId: t.Optional[str]
    """ID of an account this chatbot was created under, if applicable."""

    actions: t.List[AgentAction]
    """The unique actions this chatbot can take when interacting with users."""

    language: str = "en"
    """The language the text in this chatbot's training data is written in. Defaults to English."""

    intents: t.List[Intent]
    """The unique intents this chatbot can recognize in the things users say to it."""

    tagTypes: t.List[str]
    """The custom named entities this chatbot can recognize in a user's speech."""

    slots: t.List[Slot]
    """The potential values this chatbot can track and store over the course of a conversation with a user."""

    intentExamples: t.Dict[str, t.List[NLUExample]]
    """
    The training data that can be used to teach this chatbot how to recognize the :attr:`intents` that it has defined.
    """

    intentOODExamples: t.List[str] = []
    """
    Optional examples of utterances that a user might tell this chatbot which do not belong to any of its known
    :attr:`intents`.
    """

    trainingConversations: t.List[Conversation]
    """
    Conversations which can be used to train the chatbot what next action to take, out of its possible :attr:`actions`,
    given a conversation's state so far.
    """

    def to_nlu_dataset(self, include_ood=False) -> NLUExampleDataset:
        """
        Converts the NLU examples in this config's :attr:`intentExamples`, :attr:`intentOODExamples`, and
        :attr:`trainingConversations` into an NLU dataset, for an NLU machine learning model to train on.
        """
        copy = self.copy(deep=True)
        copy.clean()
        copy.incorporate_training_conversations()
        return NLUExampleDataset.parse_obj(copy.all_nlu_examples(include_ood))

    def to_conversation_dataset(self, expand=True) -> ConversationDataset:
        """
        Converts this config's :attr:`trainingConversations` into a conversation dataset, for a dialogue poliy
        machine learning model to train on.
        """
        copy = self.copy(deep=True)
        copy.clean()
        return ConversationDataset.from_conversations(copy.trainingConversations, expand)

    def all_nlu_examples(self, include_ood=False) -> t.List[NLUExample]:
        examples = list(chain.from_iterable(self.intentExamples.values()))
        if include_ood:
            examples += [NLUExample(text=text, isOOD=True) for text in self.intentOODExamples]
        return examples

    def intent_names(self) -> t.Set[str]:
        return set(intent.name for intent in self.intents)

    def tag_names(self) -> t.Set[str]:
        return set(self.tagTypes)

    def action_names(self) -> t.Set[str]:
        return set(a.name for a in self.actions)

    def clean(self):
        """Filters out invalid and unusable training data from the config."""
        self.filter_invalid_intent_examples()
        self.filter_no_agent_convs()
        self.filter_invalid_intent_convs()
        self.filter_invalid_action_convs()

    def filter_invalid_intent_examples(self):
        """
        Filters out all of the chatbot's NLU examples whose intents are not explicitly defined in its :attr:`intents`,
        or whose tags are not explicitly defined in its :attr:`tagTypes`.
        """
        filtered = defaultdict(list)
        valid_intents, valid_tag_types = self.intent_names(), self.tag_names()
        invalid_intents, invalid_tag_types = set(), set()
        lost_to_intents, lost_to_tags = 0, 0

        for example in self.all_nlu_examples():
            example_tag_types = set(tag.tagType for tag in example.tags or [])
            if example.intent not in valid_intents:
                invalid_intents.add(example.intent)
                lost_to_intents += 1
                continue
            if not example_tag_types.issubset(valid_tag_types):
                invalid_tag_types.update(example_tag_types - valid_tag_types)
                lost_to_tags += 1
                continue
            filtered[example.intent].append(example)

        self._warn_lost_data("NLU examples", lost_to_intents, "intents", invalid_intents)
        self._warn_lost_data("NLU examples", lost_to_tags, "tag types", invalid_tag_types)
        self.intentExamples = filtered

    def filter_no_agent_convs(self):
        """Removes all training conversations from this chatbot config which have no agent turns."""
        self.trainingConversations = [c for c in self.trainingConversations if c.num_agent_turns > 0]

    def filter_invalid_intent_convs(self):
        """Removes all training conversations that use any intents not defined in the chatbot's :attr:`intents`."""
        valid_intents = set(intent.name for intent in self.intents)
        invalid_intents = set()
        new_convs = []
        num_lost = 0
        for c in self.trainingConversations:
            if c.intents_used.issubset(valid_intents):
                new_convs.append(c)
            else:
                num_lost += 1
                invalid_intents.update(c.intents_used - valid_intents)
        self._warn_lost_data("training conversations", num_lost, "intents", invalid_intents)
        self.trainingConversations = new_convs

    def filter_invalid_action_convs(self):
        """
        Removes all training conversations that use any agent actions not defined in the chatbot's :attr:`actions`.
        """
        valid_actions = set(a.name for a in self.actions)
        invalid_actions = set()
        new_convs = []
        num_lost = 0
        for c in self.trainingConversations:
            if c.actions_used.issubset(valid_actions):
                new_convs.append(c)
            else:
                num_lost += 1
                invalid_actions.update(c.actions_used - valid_actions)
        self._warn_lost_data("training conversations", num_lost, "actions", invalid_actions)
        self.trainingConversations = new_convs

    def incorporate_training_conversations(self):
        """Adds to this agent's NLU examples any valid examples present in its training conversations."""
        valid_intents = self.intent_names()
        for conv in self.trainingConversations:
            for turn in conv.turns:
                if isinstance(turn, UserDialogueTurn):
                    if turn.userAction.intent in valid_intents and turn.userAction.utterance:
                        self.intentExamples[turn.userAction.intent].append(
                            NLUExample(intent=turn.userAction.intent, text=turn.userAction.utterance, tags=[])
                        )

    @classmethod
    def from_conversation_dataset(cls, convs: ConversationDataset, name: str, **kwargs):
        """Builds an agent config from a conversation dataset, including all its conversations and NLU examples."""
        nlu_examples = convs.to_nlu_dataset()
        examples_by_intent = defaultdict(list)
        for ex in nlu_examples:
            if not ex.isOOD and ex.intent is not None:
                examples_by_intent[ex.intent].append(ex)
        return cls(
            name=name,
            actions=[action.copy(deep=True) for action in convs.unique_actions().values()],
            intents=[Intent(name=intent) for intent in convs.unique_intents()],
            tagTypes=list(convs.unique_tag_types()),
            slots=[Slot(name=slot) for slot in convs.unique_slots()],
            intentOODExamples=list({ex.text for ex in nlu_examples if ex.isOOD}),
            intentExamples=examples_by_intent,
            trainingConversations=list(convs),
            **kwargs,
        )

    def _warn_lost_data(self, lost_type: str, num_lost: int, invalid_type: str, invalid: set):
        if num_lost > 0:
            logger.warning(
                f"There are {len(invalid)} {invalid_type} found in {lost_type} that are not defined "
                f"in the {self.name} config. Because of this, {num_lost} {lost_type} are being filtered out. "
                f"Consider either adding these {invalid_type} to your config or removing use of them from your "
                f"{lost_type}: {invalid}"
            )


class AgentExport(BaseModel):
    """
    Wrapper data structure for an :class:`AgentConfig` object. Often, agent configs are saved as JSON files in this
    format.
    """

    config: AgentConfig



import os
import typing as t

from lifeomic_chatbot_tools._utils import ImportExtraError


try:
    import boto3
    from botocore import exceptions as boto3_exceptions
    from botocore.config import Config
except ImportError:
    raise ImportExtraError("aws", __name__)


class S3Client:
    def __init__(self):
        endpoint = os.getenv("AWS_ENDPOINT")
        region = os.getenv("AWS_REGION")
        self.resource = boto3.resource("s3", endpoint_url=endpoint, config=Config(region_name=region))
        self.client = boto3.client("s3", endpoint_url=endpoint, config=Config(region_name=region))

    def upload_dir(self, source_path: str, target_bucket: str, target_path: str):
        """
        Recursively uploads all files in the directory at ``source_path`` as objects in the AWS ``target_bucket``, all
        saved under the "directory" ``target_path``.
        """
        assert os.path.isdir(source_path)
        for root, _, filenames in os.walk(source_path):
            for filename in filenames:
                file_path_local = os.path.join(root, filename)
                object_key = os.path.join(target_path, file_path_local[1 + len(source_path) :])
                self.client.upload_file(file_path_local, target_bucket, object_key)

    def download_dir(self, source_bucket: str, source_path: str, target_path: str):
        """
        Recursively downloads all files living under the ``source_path`` directory in the AWS ``source_bucket``.
        Downloads them to the local directory ``target_path``.
        """
        objects = self.resource.Bucket(source_bucket).objects.filter(Prefix=source_path)
        for obj in objects:
            obj_save_path = os.path.join(target_path, obj.key[1 + len(source_path) :])
            save_dir = os.path.abspath(os.path.dirname(obj_save_path))
            if not os.path.isdir(save_dir):
                os.makedirs(save_dir)
            os.makedirs(os.path.dirname(obj_save_path), exist_ok=True)
            self.client.download_file(source_bucket, obj.key, obj_save_path)

    def get_object(self, bucket: str, key: str, **kwargs) -> t.Optional[bytes]:
        """
        Retrieves the object living at `s3://{bucket}/{key}`. Returns `None` if there is no object living there, rather
        than throwing an error.
        """
        try:
            data = self.client.get_object(Bucket=bucket, Key=key, **kwargs)
            return data["Body"].read()
        except boto3_exceptions.ClientError as error:
            if error.response["Error"]["Code"] != "NoSuchKey":
                raise error
        return None

    def delete_object(self, bucket: str, key: str, **kwargs) -> bool:
        """
        Attempts to delete the object living at `s3://{bucket}/{key}`. Returns `True` if it was deleted, and `False` if
        no object lived there.
        """
        try:
            self.client.head_object(Bucket=bucket, Key=key)
        except boto3_exceptions.ClientError as error:
            if error.response["Error"]["Code"] != "404":
                raise error
            # The object does not exist
            return False
        self.client.delete_object(Bucket=bucket, Key=key, **kwargs)
        return True



import time
import typing as t


_T = t.TypeVar("_T")


def wait_until(condition: t.Callable[[], _T], timeout: float, check_interval: float) -> _T:
    """
    Checks the return value of ``condition``, retrying every ``check_interval`` seconds until it returns a truthy value.
    Once that happens, returns ``condition``'s return value. ``condition`` should not take any arguments. If ``timeout``
    is reached before the condition is met, a ``TimeoutError`` is raised.
    """
    start_time = time.time()
    while True:
        state = condition()
        if state:
            return state
        if time.time() - start_time > timeout:
            raise TimeoutError("condition was not met before the timeout")
        time.sleep(check_interval)



import typing as t

from lifeomic_chatbot_tools.aws.s3 import S3Client
from lifeomic_chatbot_tools.persistence.record_store.base import IRecord, IRecordStore
from lifeomic_chatbot_tools.persistence.record_store.dynamodb import DynamoDBRecordStore, KeyDict, TIDynamoDBRecord


TRecord = t.TypeVar("TRecord")


class ILargeRecord(IRecord, t.Protocol[TIDynamoDBRecord]):
    """
    Represents a database record whose payload is persisted in S3. This allows for very large records. The id
    and queryable attributes are stored in DynamoDB, but the payload is persisted in S3.
    """

    def get_attrs(self) -> TIDynamoDBRecord:
        """Gets this record's queryable attributes sub-record, which should have the same id as this record."""
        ...

    def get_body(self) -> bytes:
        """Gets the potentially large body of the record."""
        ...

    @classmethod
    def attrs_id_to_body_id(cls, attrs_id: KeyDict) -> str:
        """Converts the DynamoDB attributes id to the format used for storing the record body under in S3."""
        ...

    @classmethod
    def body_id_to_attrs_id(cls, body_id: str) -> KeyDict:
        """Converts the S3-formatted record id to the DynamoDB style used for storing the record attributes."""
        ...

    @classmethod
    def reconstitute(cls, attrs: TIDynamoDBRecord, body: bytes) -> "ILargeRecord":
        """Reconstitutes a record from its separated attributes and body."""
        ...


TILargeRecord = t.TypeVar("TILargeRecord", bound=ILargeRecord)


class ConsistencyError(Exception):
    """Raised when a data consistency error occurs in a LargeRecordStore operation."""

    pass


class LargeRecordStore(t.Generic[TILargeRecord, TIDynamoDBRecord], IRecordStore[str, TILargeRecord]):
    """
    A record store, backed by DynamoDB and S3, which supports very large records. The queryable attributes are stored
    in a pointer object in DynamoDB, and the actual record body is stored in a linked S3 object.
    """

    def __init__(
        self,
        bucket_name: str,
        record_class: t.Type[TILargeRecord],
        attrs_store: DynamoDBRecordStore[TIDynamoDBRecord],
    ):
        self._record_class = record_class
        self._attrs_store = attrs_store
        self._s3 = S3Client()
        self._bucket = bucket_name

    def save(self, record: TILargeRecord):
        """Persist a potentially large record, saving its queryable attributes in DynamoDB, and its body to S3."""
        # Save the queryable attributes.
        self._attrs_store.save(record.get_attrs())
        # Save the potentially large payload.
        self._s3.client.put_object(Bucket=self._bucket, Body=record.get_body(), Key=record.get_id())

    def get(self, id_: str) -> t.Optional[TILargeRecord]:
        """Gets the components of a record, reconstitutes it, then returns the record."""
        attrs_id = self._record_class.body_id_to_attrs_id(id_)
        attrs = self._attrs_store.get(attrs_id)
        body = self._s3.get_object(self._bucket, id_)
        if attrs is None and body is None:
            return None
        if attrs is not None and body is not None:
            return t.cast(TILargeRecord, self._record_class.reconstitute(attrs, body))
        raise ConsistencyError(
            f"during get, attrs ({attrs_id}) exists == {attrs is not None} "
            f"while body ({id_}) exists == {body is not None}"
        )

    def delete(self, id_: str) -> bool:
        """
        Deletes all components of a record. Returns `True` if both components were deleted, and `False` if one or
        more of them did not exist at the time of deletion. Raises an error if one or more deletions did not complete
        successfully.
        """
        attrs_id = self._record_class.body_id_to_attrs_id(id_)
        attrs_deleted = self._attrs_store.delete(attrs_id)
        body_deleted = self._s3.delete_object(self._bucket, id_)
        if attrs_deleted != body_deleted:
            raise ConsistencyError(
                f"during delete, attrs ({attrs_id}) exists == {attrs_deleted} "
                f"while body ({id_}) exists == {body_deleted}"
            )
        return attrs_deleted and body_deleted

    def get_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> t.Iterable[TILargeRecord]:
        """Get all chunks satisfying the condition, then de-chunk into single records."""
        for attrs in self._attrs_store.get_all(*conditions, **where_equals):
            attrs_id = attrs.get_id()
            body_id = self._record_class.attrs_id_to_body_id(attrs_id)
            body = self._s3.get_object(self._bucket, body_id)
            if body is None:
                raise ConsistencyError(
                    f"during get_all, attrs ({attrs_id}) exists == True while body ({body_id}) exists == False"
                )
            yield t.cast(TILargeRecord, self._record_class.reconstitute(attrs, body))

    def delete_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> int:
        n_deleted = 0
        for attrs in self._attrs_store.get_all(*conditions, **where_equals):
            n_deleted += int(self.delete(self._record_class.attrs_id_to_body_id(attrs.get_id())))
        return n_deleted



import typing as t

from pydantic import BaseModel

from lifeomic_chatbot_tools.ml.dataset import LabeledDataset


class TrainingTag(BaseModel):
    """A tag as it appears in NLU training data."""

    tagType: str
    """The type of named entity that is referenced e.g. ``"person"``, ``"location"``, ``"size"``, etc."""

    start: int
    """The starting index where this tag can be found in its parent utterance (inclusive)."""

    end: int
    """The ending index where this tag can be found in its parent utterance (exclusive)."""


class NLUExample(BaseModel):
    intent: t.Optional[str]
    """The intent of :attr:`text`. When ``text=="Hello, how are you?"``, the intent might be ``"greet"``."""

    text: str
    """The text of the example, e.g. "Hello, how are you?\""""

    tags: t.Optional[t.List[TrainingTag]]
    """Any tags that are present in :attr:`text`."""

    isOOD = False
    """``True`` if :attr:`text` does not belong to any intent in this example's dataset."""


class NLUExampleDataset(LabeledDataset[NLUExample, t.Optional[str]]):
    """A dataset of :class:`NLUExample` objects."""

    def get_label(self, item: NLUExample) -> t.Optional[str]:
        """Returns the intent of ``item``, an :class:`NLUExample`."""
        return item.intent

    def unique_tag_types(self) -> t.Set[str]:
        return set(tag.tagType for ex in self for tag in ex.tags or [])



import typing as t


try:
    import numpy as np
    from sklearn.model_selection import StratifiedKFold
    from sklearn.utils import shuffle as do_shuffle
except ImportError:
    _has_ml_deps = False
else:
    _has_ml_deps = True

from lifeomic_chatbot_tools._utils import requires_extras


_T = t.TypeVar("_T")


def leave_one_out(items: t.List[_T]) -> t.Iterable[t.Tuple[_T, t.List[_T]]]:
    """
    Performs leave-one-out cross validation. Cycles through ``items``. On each ith item ``item_i``, it yields
    ``item_i``, as well as all items in ``items`` except ``item_i`` as a list. So given ``items==[1,2,3,4]``, the first
    iteration would yield ``(1, [2,3,4])``, the second will yield ``(2, [1,3,4])``, and so on.
    """
    for i, item in enumerate(items):
        yield item, items[:i] + items[i + 1 :]


@requires_extras(ml=_has_ml_deps)
def make_stratified_folds(
    data: t.Sequence[_T], labels: t.Sequence, nfolds: int, shuffle: bool = True, seed: int = 0
) -> t.Tuple[t.List[_T], ...]:
    """
    Takes ``data``, a list, and breaks it into ``nfolds`` chunks. Each chunk is stratified
    by `labels`.
    """
    skf = StratifiedKFold(n_splits=nfolds, shuffle=shuffle, random_state=seed)
    fold_indices = [indices for _, indices in skf.split(data, labels)]
    folds = tuple([data[i] for i in indices] for indices in fold_indices)
    if shuffle:
        folds = tuple(do_shuffle(fold, random_state=seed) for fold in folds)
    return folds


@requires_extras(ml=_has_ml_deps)
def aggregate_dicts(dicts: t.Sequence[dict], agg: str = "mean") -> dict:
    """
    Aggregates a list of dictionaries into a single dictionary. All dictionaries in ``dicts`` should have the same keys.
    All values for a given key are aggregated into a single value using ``agg``. Returns a single dictionary with the
    aggregated values.

    Parameters
    ----------
    dicts : sequence of dicts
        The dictionaries to aggregate.
    agg : {'mean', 'stdev', 'sum', 'median', 'min', 'max'}
        Name of the method to use to aggregate the values of `dicts` with.
    """
    aggs: t.Dict[str, t.Callable] = {
        "mean": np.mean,
        "stdev": np.std,
        "sum": np.sum,
        "median": np.median,
        "min": np.min,
        "max": np.max,
    }
    assert len(dicts) > 0
    keys = dicts[0].keys()
    result = {}
    for key in keys:
        values = [d[key] for d in dicts]
        if isinstance(values[0], dict):
            # Recurse
            result[key] = aggregate_dicts(values, agg)
        else:
            result[key] = aggs[agg](values, axis=0)
    return result


@requires_extras(ml=_has_ml_deps)
def onehot(data, axis=-1, dtype=None):
    """A pure numpy implementation of the one-hot encoding function for arrays of arbitrary dimensionality.
    (`Source <https://stackoverflow.com/a/63840293>`_).
    """
    if dtype is None:
        dtype = np.float32
    pos = axis if axis >= 0 else data.ndim + axis + 1
    shape = list(data.shape)
    shape.insert(pos, data.max() + 1)
    out = np.zeros(shape, dtype)
    ind = list(np.indices(data.shape, sparse=True))
    ind.insert(pos, data)
    out[tuple(ind)] = True
    return out



import typing as t

from lifeomic_chatbot_tools.persistence.record_store.base import IRecordStore, TIRecord, operators


THashable = t.TypeVar("THashable", bound=t.Hashable)


class InMemoryRecordStore(t.Generic[THashable, TIRecord], IRecordStore[THashable, TIRecord]):
    r"""
    A simple in-memory DAO for Pydantic data models. Useful for testing or other lightweight needs. Does not keep any
    indexes of non-primary key fields, so `WHERE` clause queries are :math:`\mathcal{O}(n)`.
    """

    def __init__(self):
        # Records in the db can be resolved via `self._db[id]`.
        self._db: t.Dict[THashable, TIRecord] = {}

    def save(self, record: TIRecord):
        self._db[record.get_id()] = record

    def get(self, id_: THashable) -> t.Optional[TIRecord]:
        return self._db.get(id_)

    def delete(self, id_: THashable) -> bool:
        if self.get(id_) is not None:
            del self._db[id_]
            return True
        return False

    def get_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> t.Iterable[TIRecord]:
        for record in self._db.values():
            if self._matches(record, *conditions, **where_equals):
                yield record

    def delete_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> int:
        to_delete = [id_ for id_, record in self._db.items() if self._matches(record, *conditions, **where_equals)]
        for id_ in to_delete:
            del self._db[id_]
        return len(to_delete)

    @staticmethod
    def _matches(record: TIRecord, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> bool:
        all_conditions = list(conditions) + [(k, "==", v) for k, v in where_equals.items()]
        return all(operators[op](getattr(record, k), v) for k, op, v in all_conditions)



"""
Features related to working with numpy data in data structures and persistence. Requires the ``ml`` extra to be
installed, which can be installed in this way: ::

   pip install lifeomic-chatbot-tools[ml]

"""
import inspect
import typing as t
from io import BytesIO

from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel, root_validator

from lifeomic_chatbot_tools._utils import ImportExtraError


try:
    import numpy as np
except ImportError:
    raise ImportExtraError("ml", __name__)


def encode_numpy(data: np.ndarray, mode="w"):
    """
    Serializes a numpy array to ``bytes`` or `str`, depending on ``mode``. Includes shape, datatype, and endianness
    information for perfect cross-platform reconstruction.

    Parameters
    ----------
    data : np.ndarray
        The data to serialize.
    mode : {'w', 'wb'}
        The serialization mode to use. if ``mode=="w"``, the output will be a string. If ``mode=="wb"``, the output
        will be ``bytes``.
    """
    mf = BytesIO()
    np.save(mf, data)
    value = mf.getvalue()
    mf.close()
    if mode == "w":
        return value.decode("latin-1")
    return value


def decode_numpy(data: t.Union[str, bytes]):
    """
    Deserializes a numpy array from ``bytes`` or ``str``, which was serialized using the :func:`encode_numpy` method.
    """
    if isinstance(data, str):
        data = data.encode("latin-1")
    mf = BytesIO(data)
    arr = np.load(mf)
    mf.close()
    return arr


class DataModel(BaseModel):
    """
    A base class for defining pydantic models which automatically support numpy fields, including serialization and
    deserialization of those fields. E.g.

    >>> import numpy as np
    ...
    ... class MyModel(DataModel):
    ...     a: str
    ...     b: np.ndarray
    ...
    ... model = MyModel(a="hello world", b=np.array([1,2,3]))
    ... # Serialize to a string and then back again.
    ... reconstructed = MyModel.parse_raw(model.json())
    ... # `reconstructed`'s contents are identical to `model`
    Serializing to primitive python data structures is also supported, via :meth:`dict` and :meth:`parse_obj`.
    """

    class Config:
        arbitrary_types_allowed = True

    def dict(self, custom_encoder: t.Optional[t.Dict[t.Any, t.Callable[[t.Any], t.Any]]] = None, **kwargs):
        """
        Creates a dictionary representation of the model, encoding all values to basic Python data types, for example,
        enum values become strings, and numpy arrays become data strings.
        """
        _custom_encoder = {np.ndarray: lambda arr: encode_numpy(arr, mode="w")}
        if custom_encoder is not None:
            # Accept user overrides and customization.
            _custom_encoder.update(custom_encoder)
        # First convert to a `dict`, to prevent `jsonable_encoder` from recursively calling this object's `dict` method,
        # which would cause infinite recursion.
        d = super().dict(**kwargs)
        # TODO: the top level call to `DataModel.json()` bloats the string returned here by `encode_numpy`,
        #   because it inserts escape characters and converts '\x' to '\u00' sometimes.
        return jsonable_encoder(d, custom_encoder=_custom_encoder, **kwargs)

    @classmethod
    def _get_fields_of_type(cls, type_: t.Type) -> t.Set[str]:
        """Get the name of the fields in this pydantic model that have an annotation of `type_`."""
        sig = inspect.signature(cls)
        return {param.name for param in sig.parameters.values() if param.annotation == type_}

    @root_validator(pre=True)
    def _validate_numpy_arrays(cls, values):
        """
        Allows numpy data to be passed in to this model's constructor in a few different forms, namely:
            - A normal `np.ndarray` object.
            - A `str` object, assumed to be produced by `encode_numpy`. It will be decoded.
            - A 'bytes` object, assumed to be produced by `encode_numpy`. It will be decoded.
        """
        numpy_fields = cls._get_fields_of_type(np.ndarray)
        for field in numpy_fields:
            value = values[field]
            if isinstance(value, np.ndarray):
                continue
            elif isinstance(value, (str, bytes)):
                values[field] = decode_numpy(value)
            else:
                raise TypeError(f"cannot process unknown type {type(value)} for {field} field")
        return values



from pydantic import BaseModel


class StrPredWithConf(BaseModel):
    """
    Represents a string prediction that was made (:attr:`value`), along with the :attr:`confidence` of that prediction.
    """

    value: str
    confidence: float



import typing as t

from pydantic import BaseModel, Field

from lifeomic_chatbot_tools.types.conversations.actions import Actor, AgentAction, HumanAgentAction, UserAction


class DialogueState(BaseModel):
    slotValues: t.Dict[str, t.Any]


class BaseDialogueTurn(BaseModel):
    state: t.Optional[DialogueState]
    actor: Actor


class UserDialogueTurn(BaseDialogueTurn):
    userAction: UserAction
    actor = Field(Actor.USER, const=True)


class AgentDialogueTurn(BaseDialogueTurn):
    agentAction: AgentAction
    actor = Field(Actor.AGENT, const=True)


class HumanAgentDialogueTurn(BaseDialogueTurn):
    humanAgentAction: HumanAgentAction
    actor = Field(Actor.HUMAN_AGENT, const=True)


DialogueTurn = t.Union[AgentDialogueTurn, HumanAgentDialogueTurn, UserDialogueTurn]



import json
import typing as t
from json import JSONDecodeError

import requests
from pydantic import BaseModel, ValidationError

from lifeomic_chatbot_tools._utils import ImportExtraError


try:
    import boto3
except ImportError:
    raise ImportExtraError("aws", __name__)


class AlphaResponse(BaseModel):
    status_code: int  # the http response status code
    text: str  # the http response body

    @property
    def body(self):
        """Attempts to parse the response body as JSON."""
        try:
            return json.loads(self.text)
        except JSONDecodeError as e:
            raise RuntimeError(f"could not parse text {self.text} as json, reason: {e}")

    @property
    def ok(self):
        """Returns ``True`` if the response's status code is in the 200-300 range."""
        return self.status_code < 400


class Alpha:
    """
    A minimal Python port of LifeOmic's `alpha` utility for calling Lambda functions that operate
    as web services using the [AWS API Gateway event format](https://docs.aws.amazon.com/lambda/latest/dg/services-apiga
    teway.html#apigateway-example-event).
    """

    def __init__(self, target: str):
        """
        If ``target`` begins with ``lambda://`` e.g. ``lambda://function-name``, then ``boto3`` will attempt to use the
        environment credentials and call an actual Lambda function named ``function-name``. Alternatively, an actual URL
        can be passed in as the ``target`` to support calling e.g. a locally running Lambda function.
        """
        self._target = target
        self._client = None
        prefix = "lambda://"
        if target.startswith(prefix):
            self._target = target[len(prefix) :]
            self._client = boto3.client("lambda")

    def get(self, path: str, params: t.Optional[t.Dict[str, str]] = None, headers: t.Optional[t.Dict[str, str]] = None):
        payload = self._make_payload(path=path, method="GET", params=params, headers=headers)
        return self._invoke_lambda(payload)

    def post(self, path: str, body: t.Optional[t.Dict[str, str]] = None, headers: t.Optional[t.Dict[str, str]] = None):
        payload = self._make_payload(path=path, method="POST", body=body, headers=headers)
        return self._invoke_lambda(payload)

    def put(self, path: str, body: t.Optional[t.Dict[str, str]] = None, headers: t.Optional[t.Dict[str, str]] = None):
        payload = self._make_payload(path=path, method="PUT", body=body, headers=headers)
        return self._invoke_lambda(payload)

    def delete(self, path: str, headers: t.Optional[t.Dict[str, str]] = None):
        payload = self._make_payload(path=path, method="DELETE", headers=headers)
        return self._invoke_lambda(payload)

    @staticmethod
    def _make_payload(
        path: str,
        method: str,
        body: t.Optional[t.Dict[str, str]] = None,
        params: t.Optional[t.Dict[str, str]] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
    ):
        payload: t.Dict[str, t.Union[str, t.Dict[str, str]]] = {"path": path, "httpMethod": method}
        if body:
            payload["body"] = json.dumps(body)
        if params:
            payload["queryStringParameters"] = params
        if headers:
            payload["headers"] = headers
        return payload

    def _invoke_lambda(self, payload: dict):
        if self._client:
            res = self._client.invoke(FunctionName=self._target, Payload=json.dumps(payload))
            return self._parse_response(res["Payload"].read())
        else:
            res = requests.post(self._target, json=payload)
            return self._parse_response(res.content)

    @staticmethod
    def _parse_response(payload: bytes):
        """Creates an `AlphaResponse` object from a raw Lambda response payload."""
        try:
            parsed = json.loads(payload.decode("utf-8"))
            return AlphaResponse(status_code=parsed["statusCode"], text=parsed["body"])
        except (JSONDecodeError, KeyError, ValidationError) as e:
            raise RuntimeError(f"could not parse payload {payload!r} as an API Gateway event, reason: {e}")



import typing as t
from collections import defaultdict
from statistics import mean

from lifeomic_chatbot_tools._utils import ImportExtraError


try:
    import networkx as nx
except ImportError:
    raise ImportExtraError("ml", __name__)

from lifeomic_chatbot_tools.types.conversations.actions import Actor
from lifeomic_chatbot_tools.types.conversations.conversation import ConversationDataset
from lifeomic_chatbot_tools.types.conversations.dialogue_turns import (
    AgentDialogueTurn,
    DialogueTurn,
    HumanAgentDialogueTurn,
)


class DSTuple(t.NamedTuple):
    """A dialogue state, as defined in [1]_. It represents the state of a conversation at a given turn."""

    bs: t.FrozenSet[str] = frozenset()
    """The set of dialogue slots that are populated."""

    action: t.Optional[str] = None
    """The action that was taken."""

    actor: t.Optional[str] = None
    """
    The actor that took the action, one of ``"HUMAN"``, ``"AGENT"`` (chatbot), or ``None`` (in the case of the
    starting root node of the graph).
    """


class ConvGraph:
    """
    Implementation of Algorithm (1) in [1]_. Can convert a
    :class:`~lifeomic_chatbot_tools.types.conversations.conversation.ConversationDataset` into a conversation graph,
    which is useful for visualizing the different paths that an agent's conversations take, as well as for evaluating
    dialogue act predictions using "soft" metrics. The graph can also be used for dataset augmentation in a
    task-oriented-dialogue training setting.

    Each node in the graph is a unique dialogue state, where a dialogue state is defined as a tuple of:

    #. the current belief state
    #. the action that was taken
    #. the type of actor who took that action (human or chatbot).

    A belief state is the set of dialogue slots that are populated at a given turn.

    References
    ----------
    .. [1] Gritta, M., Lampouras, G., & Iacobacci, I. (2021). Conversation Graph: Data Augmentation, Training, and
       Evaluation for Non-Deterministic Dialogue Management. Transactions of the Association for Computational
       Linguistics, 9, 36-52.
    """

    def __init__(self, data: ConversationDataset):
        self.graph = nx.DiGraph()
        for conv in data:
            last_state = DSTuple()  # the starting node of the graph
            self.add_node(last_state)
            for turn in conv.turns:
                if turn.actor == Actor.HUMAN_AGENT:
                    continue
                ds = self.encode_dialogue_state(turn)
                self.add_node(ds)
                if not self.graph.has_edge(last_state, ds):
                    self.graph.add_edge(last_state, ds, weight=1)
                else:
                    self.graph[last_state][ds]["weight"] += 1
                last_state = ds

    def add_node(self, ds: DSTuple):
        """
        The ``ds`` tuple is used as the node's unique id, but we also add the attributes of ``ds`` as attributes of the
        node, so they can be interacted with easily in the graph.
        """
        if not self.graph.has_node(ds):
            bs_str = ", ".join(sorted(ds.bs))
            self.graph.add_node(ds, bs=bs_str, action=ds.action, actor=ds.actor)

    def soft_accuracy(self, y_pred: t.List[str], last_turns: t.List[DialogueTurn]) -> float:
        """
        Calcuates the soft accuracy, which is accuracy when there is more than one acceptable answer for a prediction.

        Parameters
        ----------
        y_pred : list of str
            A list of agent action predictions.
        last_turns : list of DialogueTurn
            A list, having the same length as ``y_pred``, of the dialogue turns the agent action predictions in
            ``y_pred`` are following. For example, ``y_pred[i]`` should be the agent action which was predicted as
            coming after the turn ``last_turns[i]``.
        """
        num_correct = 0
        for pred, last_turn in zip(y_pred, last_turns):
            num_correct += self.is_pred_correct(pred, last_turn)
        return num_correct / len(last_turns)

    def balanced_soft_accuracy(self, y_pred: t.List[str], last_turns: t.List[DialogueTurn]) -> float:
        """Same as :meth:`soft_accuracy`, but equally weights the accuracy calculation of each class, or action."""
        correct: t.Dict[str, int] = defaultdict(int)
        out_of: t.Dict[str, float] = defaultdict(int)
        for pred, last_turn in zip(y_pred, last_turns):
            valid_next_actions = self.get_valid_next_actions(last_turn)
            if pred in valid_next_actions:
                # The prediction was right, so give full credit.
                correct[pred] += 1
                out_of[pred] += 1
            else:
                # The prediction was wrong, so equally penalize each of the classes that were valid.
                for label in valid_next_actions:
                    out_of[label] += 1 / len(valid_next_actions)
        # Compute the soft accuracy for each class/action, then return the mean of those.
        return mean(correct[action] / out_of[action] for action in out_of.keys())

    def is_pred_correct(self, pred: str, last_turn: DialogueTurn) -> bool:
        valid_next_actions = self.get_valid_next_actions(last_turn)
        return pred in valid_next_actions

    def get_valid_next_actions(self, turn: DialogueTurn) -> t.Set[str]:
        """
        Given the dialogue state of ``turn``, find the actions of all dialogue states that directly follow ``turn`` in
        the conversation graph. In other words, return the valid next actions that appear after `turn` in the training
        data the conversation graph was constructed from.
        """
        ds = self.encode_dialogue_state(turn)
        return {v[1] for _, v in self.graph.out_edges(ds)}  # type: ignore

    def avg_num_valid_next_actions(self) -> float:
        """
        The average out-degree of the conversation graph, which is equal to the average number of correct agent actions
        for each dialogue state in the graph.
        """
        return mean(n for _, n in self.graph.out_degree)

    def save(self, path: str):
        nx.write_graphml(self.graph, path)

    @staticmethod
    def encode_dialogue_state(turn: DialogueTurn) -> DSTuple:
        """
        Encodes a :obj:`~lifeomic_chatbot_tools.types.conversations.dialogue_turns.DialogueTurn` into a
        :class:`DSTuple`.
        """
        if isinstance(turn, HumanAgentDialogueTurn):
            raise AssertionError("cannot compute dialogue state for a human agent turn")
        belief_state_keys = list(turn.state.slotValues.keys()) if turn.state else []
        action = turn.agentAction.name if isinstance(turn, AgentDialogueTurn) else turn.userAction.intent
        return DSTuple(bs=frozenset(belief_state_keys), action=action, actor=turn.actor.value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lifeomic_chatbot_tools',
 'lifeomic_chatbot_tools.aws',
 'lifeomic_chatbot_tools.ml',
 'lifeomic_chatbot_tools.persistence',
 'lifeomic_chatbot_tools.persistence.record_store',
 'lifeomic_chatbot_tools.types',
 'lifeomic_chatbot_tools.types.conversations',
 'lifeomic_chatbot_tools.web']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.63.0,<1.0.0',
 'loguru>=0.5.1,<1.0.0',
 'pydantic>=1.7.3,<1.9.0',
 'requests>=2.21.0,<3.0.0']

extras_require = \
{'aws': ['boto3>=1.21.7,<2.0.0'],
 'ml': ['numpy>=1.19.2,<2.0.0',
        'scikit-learn>=0.24.2,<2.0.0',
        'networkx>=2.5.1,<3.0.0']}

setup_kwargs = {
    'name': 'lifeomic-chatbot-tools',
    'version': '5.0.0',
    'description': 'Utilities for machine learning, web services, and cloud infrastructure.',
    'long_description': '# lifeomic-chatbot-tools\n\nPython utilities for machine learning, web services, and cloud infrastructure.\nIncludes classes and methods for:\n\n1. ML model serialization/deserialization\n2. ML model evaluation utilities\n3. Data structures/models related to chatbots\n4. ML model artifact persistence and version management\n5. And more\n\nThe data structures in this package can all be found in the\n`lifeomic_chatbot_tools.types` sub-package, and are all\n[Pydantic](https://pydantic-docs.helpmanual.io/) data models. For example the\n`lifeomic_chatbot_tools.types.agent.AgentConfig` class represents a chatbot\'s\nconfiguration and training data.\n\n## Getting Started\n\nTo begin using the package, use your favorite package manager to install it from PyPi.\nFor example, using pip:\n\n```\npip install lifeomic-chatbot-tools\n```\n\nSome of the features in this repo require more heavy weight dependencies, like AWS\nrelated utilities, or utilities specific to machine learning. If you try to import\nthose features, they will tell you if you do not have the correct package extra\ninstalled. For example, many of the features in the `lifeomic_chatbot_tools.ml`\nsub-package require the `ml` extra. To install `lifeomic-chatbot-tools` with that\nextra:\n\n```\npip install lifeomic-chatbot-tools[ml]\n```\n\nYou can then begin using any package features that require ML dependencies.\n\n## Developing Locally\n\nBefore making any new commits or pull requests, please complete these steps.\n\n1. Install the Poetry package manager for Python if you do not already have it.\n   Installation instructions can be found\n   [here](https://python-poetry.org/docs/#installation).\n2. Clone the project.\n3. From the root directory of the repo, install the dependencies, including all dev\n   dependencies and extras:\n   ```\n   poetry install --extras "aws ml"\n   ```\n4. Install the pre-commit hooks, so they will run before each local commit. This\n   includes linting, auto-formatting, and import sorting:\n   ```\n   pre-commit install\n   ```\n   **Note**: pre-commit is a dev dependency of the project, and is managed by Poetry.\n   In order to use those dependencies from the terminal, you must be in a Poetry shell.\n   This can be accomplished by first running `poetry shell` in your terminal.\n\n## Testing Locally\n\nWith Yarn, Docker, and docker-compose installed, run this command from the project\nroot:\n\n```\nyarn ci-test\n```\n\nThis will build the project, lint it, and run the unit tests and integration tests.\nAll those steps can be run individually as well. See the scripts in the `package.json`\nfile for the command names.\n\n## Releasing The Package\n\nReleasing the package is automatically handled by CI, but three steps must be taken\nto trigger a successful release:\n\n1. Use Poetry\'s [`version` command](https://python-poetry.org/docs/cli/#version) to\n   bump the package\'s version.\n2. Commit and tag the repo with the exact same version the package was bumped to,\n   e.g. `v1.0.0` (include a preceding `v`).\n3. Push the commit and tag to remote. These can be done together using:\n   `git push --atomic origin <branch_name> <tag>`\n\nCI will then build release the package to pypi with that version once the commit and\ntag are merged to the `master` branch.\n',
    'author': 'LifeOmic Development',
    'author_email': 'development@lifeomic.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import operator
import typing as t


TId_co = t.TypeVar("TId_co", covariant=True)


class IRecord(t.Protocol[TId_co]):
    def get_id(self) -> TId_co:
        ...


TIRecord = t.TypeVar("TIRecord", bound=IRecord)
TId_contra = t.TypeVar("TId_contra", contravariant=True)
T = t.TypeVar("T")
operators = {"<=": operator.le, ">=": operator.ge, "<": operator.lt, ">": operator.gt, "==": operator.eq}


class IRecordStore(t.Protocol[TId_contra, T]):
    """Interface for a Data Access Object (DAO) which can save objects to some back-end persistence solution."""

    def save(self, record: T):
        """Updates ``record`` in the database, or creates it if it's not there."""
        ...

    def get(self, id_: TId_contra) -> t.Optional[T]:
        """Retrieves a record from the database, returning ``None`` if it doesn't exist."""
        ...

    def delete(self, id_: TId_contra) -> bool:
        """
        Deletes a record from the database, returning ``True`` if the record was deleted, and ``False`` if it didn't
        exist.
        """
        ...

    def get_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> t.Iterable[T]:
        """Retrieves all records which satisfy the optional conditions."""
        ...

    def delete_all(self, *conditions: t.Tuple[str, str, t.Any], **where_equals) -> int:
        """
        Deletes all records which satisfy the optional conditions. Returns the number of records that were deleted.
        """
        ...



import typing as t
from enum import Enum

from pydantic import BaseModel


class TagValue(BaseModel):
    """Represents a named entity's type and value."""

    tagType: str
    value: str


class Actor(Enum):
    USER = "USER"
    AGENT = "AGENT"
    HUMAN_AGENT = "HUMAN_AGENT"


class Sentiment(Enum):
    NEGATIVE = "NEGATIVE"
    NEUTRAL = "NEUTRAL"
    POSITIVE = "POSITIVE"


class UserAction(BaseModel):
    """Represents any type of user action (email, utterance, option, etc)."""

    type: str
    utterance: t.Optional[str]
    translatedUtterance: t.Optional[str]
    intent: t.Optional[str]
    sentiment: t.Optional[Sentiment]
    confidence: t.Optional[float]
    ood: t.Optional[bool]
    oodConfidence: t.Optional[float]
    tags: t.Optional[t.List[TagValue]]


class AgentAction(BaseModel):
    """Represents any type of agent action (form, utterance, email, etc.)"""

    type: str
    name: str  # the action's name
    utterance: t.Optional[str]
    requiredSlots: t.Optional[t.List[str]]


class HumanAgentAction(BaseModel):
    type: str
    utterance: t.Optional[str]



import typing as t

from pydantic import BaseModel

from lifeomic_chatbot_tools.ml.dataset import LabeledDataset
from lifeomic_chatbot_tools.types.conversations.actions import Actor, AgentAction
from lifeomic_chatbot_tools.types.conversations.dialogue_turns import AgentDialogueTurn, DialogueTurn, UserDialogueTurn
from lifeomic_chatbot_tools.types.nlu import NLUExample, NLUExampleDataset


class Conversation(BaseModel):
    id: t.Optional[str]
    agentId: t.Optional[str]
    turns: t.List[DialogueTurn]

    @property
    def num_agent_turns(self) -> int:
        return sum(1 for turn in self.turns if turn.actor == Actor.AGENT)

    @property
    def last_turn(self) -> t.Optional[DialogueTurn]:
        return None if len(self.turns) == 0 else self.turns[-1]

    @property
    def intents_used(self):
        intents = set(turn.userAction.intent for turn in self.turns if isinstance(turn, UserDialogueTurn))
        intents.discard("")
        intents.discard(None)
        return intents

    @property
    def actions_used(self):
        actions = set(turn.agentAction.name for turn in self.turns if isinstance(turn, AgentDialogueTurn))
        actions.discard("")
        return actions

    def expand(self) -> t.List["Conversation"]:
        """Turns this conversation into a list of its partial conversations that each end with an agent action."""
        cls = self.__class__
        convs = []
        for i in range(len(self)):
            if self.turns[i].actor == Actor.AGENT:
                convs.append(cls(turns=self.turns[: i + 1]))
        return convs

    def __len__(self):
        return len(self.turns)

    def get_final_user_utterance(self) -> str:
        """
        Retrieves the user utterance from the most recent
        turn of the conversation. If the most recent turn is not
        a user turn or doesn't have an utterance, returns the empty string.
        """
        last_turn = self.last_turn
        if isinstance(last_turn, UserDialogueTurn) and last_turn.userAction.type == "UTTERANCE_ACTION":
            return last_turn.userAction.translatedUtterance or last_turn.userAction.utterance or ""
        return ""


class ConversationDataset(LabeledDataset[Conversation, str]):
    def get_label(self, item: Conversation) -> str:
        """
        Returns the name of the final agent action in ``item`` (a :class:`Conversation`), and ensures the last turn
        in the conversation is in fact an agent turn.
        """
        if not isinstance(item.last_turn, AgentDialogueTurn):
            raise AssertionError("conversations in a ConversationDataset must have an agent action as final last turn.")
        return item.last_turn.agentAction.name

    @classmethod
    def from_conversations(cls, convs: t.List[Conversation], expand=True) -> "ConversationDataset":
        """
        Safely builds a dataset from Conversations which may or may not have agent actions as the final turn.
        Expands each conversation in ``convs`` into as many possible conversations as possible, under the constraint
        that each conversation end with an agent action. If ``expand==False``, expansion is bypassed, and the regular
        conversations are used.
        """
        result = []
        for conv in convs:
            if expand:
                result += conv.expand()
            else:
                result.append(conv)
        return cls.parse_obj(result)

    def turns(self) -> t.Iterable[DialogueTurn]:
        """Iterates over all turns of all conversations in the dataset."""
        for conv in self:
            for turn in conv.turns:
                yield turn

    def user_turns(self) -> t.Iterable[UserDialogueTurn]:
        for turn in self.turns():
            if isinstance(turn, UserDialogueTurn):
                yield turn

    def agent_turns(self) -> t.Iterable[AgentDialogueTurn]:
        for turn in self.turns():
            if isinstance(turn, AgentDialogueTurn):
                yield turn

    def unique_intents(self) -> t.Set[str]:
        intents = set()
        for turn in self.user_turns():
            if turn.userAction.intent is not None:
                intents.add(turn.userAction.intent)
        return intents

    def unique_actions(self) -> t.Dict[str, AgentAction]:
        """Returns a mapping of unique action names, each to an example of that action found in the dataset."""
        actions = {}
        for turn in self.agent_turns():
            actions[turn.agentAction.name] = turn.agentAction
        return actions

    def unique_tag_types(self) -> t.Set[str]:
        tag_types = set()
        for turn in self.user_turns():
            if turn.userAction.tags is not None:
                for tag in turn.userAction.tags:
                    tag_types.add(tag.tagType)
        return tag_types

    def unique_slots(self) -> t.Set[str]:
        slots = set()
        for turn in self.turns():
            if turn.state is not None:
                for slot_name in turn.state.slotValues:
                    slots.add(slot_name)
        return slots

    def to_nlu_dataset(self):
        examples = []
        for turn in self.user_turns():
            action = turn.userAction
            if action.utterance is not None:
                if action.ood:
                    examples.append(NLUExample(text=action.utterance, isOOD=True))
                elif action.intent is not None:
                    examples.append(NLUExample(text=action.utterance, intent=action.intent))
        return NLUExampleDataset.parse_obj(examples)

    def make_validation_pairs(self) -> t.Tuple["ConversationDataset", t.List[str]]:
        """
        Takes all the conversations and returns each one having all its turns but the
        last agent action. Also returns a tuple of those last agent actions.

        Returns
        -------
        tuple of (ConversationDataset, str)
            The first list is the list of conversations. The second
            is the list of the names of the next actions that should
            be taken, given the conversations; one action per conversation.
        """
        cls = self.__class__
        val_convs = []
        next_actions = []
        for conv in self:
            val_convs.append(Conversation(turns=conv.turns[:-1]))
            next_actions.append(self.get_label(conv))
        return cls.parse_obj(val_convs), next_actions



import typing as t
from abc import abstractmethod
from collections import Counter, defaultdict
from itertools import chain

from pydantic.generics import GenericModel

from lifeomic_chatbot_tools._utils import requires_extras


try:
    from sklearn.model_selection import RepeatedStratifiedKFold, train_test_split
    from sklearn.utils import resample
except ImportError:
    _has_ml_deps = False
else:
    _has_ml_deps = True


_Instance = t.TypeVar("_Instance")
_Label = t.TypeVar("_Label", bound=t.Hashable)


class LabeledDataset(GenericModel, t.Generic[_Instance, _Label]):
    """
    An abstract typed array with attached helper functions related to categorically-labeled datasets. Subclassing
    instances can be used as regular lists with indexeing, etc. The only method that needs to be implemented is
    :meth:`get_label`, which should return the label associated with an instance of this dataset.

    >>> from typing import Tuple
    ...
    ... class SentimentDataset(LabeledDataset[Tuple[str, int], int]):
    ...     def get_label(self, item):
    ...         return item[1]
    ...
    ... dataset = SentimentDataset.parse_obj(
    ...     [("I hate this", -1), ("no good", -1), ("I love this", 1), ("This is great", 1)]
    ... )
    ... dataset.labels()
    ... # [-1, 1, 1]
    ... dataset.unique_labels()
    ... # {-1, 1}
    ... train, test = next(dataset.cv(nfolds=2))
    ... # train == [('I hate this', -1), ('This is great', 1)]
    ... # test == [('no good', -1), ('I love this', 1)]
    """

    __root__: t.List[_Instance]

    @abstractmethod
    def get_label(self, item: _Instance) -> _Label:
        """
        The only method that needs to be implemented by a concrete subclass. Given an ``item`` in the dataset, returns
        the classification label for that item.
        """
        pass

    def labels(self):
        return [self.get_label(item) for item in self]

    def unique_labels(self):
        return set(self.get_label(item) for item in self)

    def get_label_distribution(self):
        """
        Counts the number of each type of label present in ``self``. The returned :class:`Counter` object can be treated
        as a dictionary e.g. ``my_label_count = counter["my_label"]``.
        """
        return Counter(self.labels())

    @requires_extras(ml=_has_ml_deps)
    def cv(
        self, nfolds: int = 5, nrepeats: int = 1, seed: int = 0
    ) -> t.Iterator[t.Tuple["LabeledDataset", "LabeledDataset"]]:
        """
        Yields train/test splits for ``nfolds`` cross-validation, stratified by label. Repeats the random splitting
        ``nrepeats`` times.
        """
        cls = self.__class__
        rskf = RepeatedStratifiedKFold(n_splits=nfolds, n_repeats=nrepeats, random_state=seed)
        for train_index, test_index in rskf.split(self, self.labels()):
            yield cls.parse_obj(self[i] for i in train_index), cls.parse_obj(self[i] for i in test_index)

    @requires_extras(ml=_has_ml_deps)
    def balance(self, seed: int = 0):
        """
        Makes a new version of ``self``, where the minority labels are upsampled to have the same number of items as the
        majority label.
        """
        cls = self.__class__
        items_by_label = defaultdict(list)
        for item in self:
            items_by_label[self.get_label(item)].append(item)
        n_majority_label = max(len(items) for items in items_by_label.values())
        upsampled = chain.from_iterable(
            resample(items, replace=True, n_samples=n_majority_label, random_state=seed)
            for items in items_by_label.values()
        )
        return cls.parse_obj(list(upsampled))

    @requires_extras(ml=_has_ml_deps)
    def split(self, test_size: t.Union[float, int, None] = None, seed: int = 0, shuffle: bool = True):
        """Returns a train/test split of ``self``, stratified by label."""
        cls = self.__class__
        train, test = train_test_split(
            self, test_size=test_size, random_state=seed, shuffle=shuffle, stratify=self.labels()
        )
        return cls.parse_obj(train), cls.parse_obj(test)

    def __iter__(self) -> t.Iterator[_Instance]:
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def __setitem__(self, key, value):
        self.__root__[key] = value

    def __len__(self):
        return len(self.__root__)
